package be.iminds.ilabt.jfed.espec.filefetcher;

import be.iminds.ilabt.jfed.espec.bundle.BundleFetcher;
import be.iminds.ilabt.jfed.espec.bundle.ESpecBundle;
import be.iminds.ilabt.jfed.espec.model.*;
import be.iminds.ilabt.jfed.espec.util.ESpecLogger;
import be.iminds.ilabt.jfed.git.GitAuthPreferences;
import be.iminds.ilabt.jfed.git.GitFetcher;
import be.iminds.ilabt.jfed.git.GitFetcherCommand;
import be.iminds.ilabt.jfed.git.GitFetcherCommandBuilder;
import be.iminds.ilabt.jfed.rspec.generator.RSpecGeneratorFactory;
import be.iminds.ilabt.jfed.util.common.IOUtils;
import be.iminds.ilabt.jfed.util.common.ThreadFactoryUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.File;
import java.io.IOException;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

/**
 * ExperimentSpecificationFileManager
 *   Fetches all files in an ExperimentSpecification, putting them in memory.
 *   It downloads any files needing download in parallell,
 *   and has a method to wait for all files to be ready.
 */
public class ExperimentSpecificationFileManager {
    private static final Logger LOG = LoggerFactory.getLogger(ExperimentSpecificationFileManager.class);

    @Nonnull private final ExperimentSpecification eSpec;
    @Nonnull private final ESpecBundle bundle;
    @Nonnull private final ESpecLogger eSpecLogger;
    @Nullable private final RSpecGeneratorFactory rSpecGeneratorFactory;
    @Nullable private final GitAuthPreferences gitAuthPreferences;

    private final Map<FileSource, byte[]> bytes = new ConcurrentHashMap<>();
    private final Set<FileSource> dirs = Collections.newSetFromMap(new ConcurrentHashMap<>());
    private final ExecutorService executorService;

    private boolean started;
    private boolean errorOccured;

    private AtomicInteger slowFetchCountdown;

    public ExperimentSpecificationFileManager(@Nonnull ExperimentSpecification eSpec,
                                              @Nonnull ESpecBundle bundle,
                                              @Nonnull ESpecLogger eSpecLogger,
                                              @Nullable RSpecGeneratorFactory rSpecGeneratorFactory,
                                              @Nullable GitAuthPreferences gitAuthPreferences) {
        this.eSpec = eSpec;
        this.bundle = bundle;
        this.eSpecLogger = eSpecLogger;
        this.rSpecGeneratorFactory = rSpecGeneratorFactory;
        this.gitAuthPreferences = gitAuthPreferences;
        this.executorService = Executors.newFixedThreadPool(3, ThreadFactoryUtil.getFactory("ESpecFileManager"));
        this.errorOccured = false;
    }

    public void fetchAll() {
        eSpecLogger.firePreFileLoad();
        started = true;

        List<FileSource> fileSources = new ArrayList<>();
        eSpec.getUploads().stream().map(UploadSpec::getSource).filter(Objects::nonNull).forEach(s -> fileSources.add(s));
        eSpec.getExecutes().stream().map(ExecuteSpec::getSource).filter(Objects::nonNull).forEach(s -> fileSources.add(s));
        if (eSpec.getAnsible() != null) {
            eSpec.getAnsible().getAnsibleHostSpec().getUploadLists().stream()
                    .map(UploadSpec::getSource).filter(Objects::nonNull).forEach(s -> fileSources.add(s));
            eSpec.getAnsible().getAnsibleHostSpec().getExecuteLists().stream()
                    .map(ExecuteSpec::getSource).filter(Objects::nonNull).forEach(s -> fileSources.add(s));
            eSpec.getAnsible().getAnsibleGalaxySpecs().stream()
                    .map(UploadLikeSpec::getSource).filter(Objects::nonNull).forEach(s -> fileSources.add(s));
            eSpec.getAnsible().getAnsiblePlaybookSpecs().stream()
                    .map(UploadLikeSpec::getSource).filter(Objects::nonNull).forEach(s -> fileSources.add(s));
            eSpec.getAnsible().getAnsiblePlaybookSpecs().stream()
                    .map(AnsiblePlaybookSpec::getExtraVarsJsonFileSpec)
                    .filter(Objects::nonNull)
                    .map(UploadLikeSpec::getSource)
                    .filter(Objects::nonNull)
                    .forEach(s -> fileSources.add(s));
        }

        Map<Boolean, List<FileFetcher>> fileSourceListBySpeed =
                fileSources.stream()
                        .map(fileSource -> new FileFetcher(fileSource, bundle, rSpecGeneratorFactory, gitAuthPreferences))
                        .collect(Collectors.partitioningBy(FileFetcher::isFast));

        List<FileFetcher> fastFetchers = fileSourceListBySpeed.get(true);
        List<FileFetcher> slowFetchers = fileSourceListBySpeed.get(false);

        for (FileFetcher fff : fastFetchers) {
            //generated keypair is handled in a different way
            if (fff.getFileSource().isGeneratedKeyPair())
                continue;
            if (fff.getFileSource().getType() == FileSource.SourceType.META)
                continue;

            try {
                if (fff.isDir()) {
                    dirs.add(fff.getFileSource());
                    continue;
                }

                byte[] b = fff.fetchBytes();
                assert b != null;
                LOG.debug("ExperimentSpecificationFileManager fetched "+b.length+" bytes for "+fff.getFileSource().getBasename());
                bytes.put(fff.getFileSource(), b);
                eSpecLogger.firePostFileLoadSuccess(fff.getFileSource(), true, b.length);
            } catch (IOException e) {
                LOG.error("Failed to (fast) fetch file with source type "+fff.getFileSource().getType()+"", e);
                errorOccured = true;
                executorService.shutdown();
                eSpecLogger.firePostFileLoadFail(fff.getFileSource(), true, 0,
                        "Failed to (fast) fetch file with source type "+fff.getFileSource().getType()+"", e);
                eSpecLogger.firePostFileLoadAll(false);
                return;
            }
        }

        LOG.debug("ExperimentSpecificationFileManager fetched "+fastFetchers.size()+" fast files.");

        if (!slowFetchers.isEmpty()) {
            slowFetchCountdown = new AtomicInteger(slowFetchers.size());
            for (FileFetcher sff : slowFetchers) {
                //generated keypair is handled differently
                if (sff.getFileSource().isGeneratedKeyPair())
                    continue;
                try {
                    if (sff.isDir()) {
                        dirs.add(sff.getFileSource());
                        continue;
                    }
                } catch (IOException e) {
                    LOG.error("Error checking if filefetch is a dir. " +
                            "Will ignore for now, and assume it is not.", e);
                }

                queueFileFetch(sff);
            }

            LOG.debug("ExperimentSpecificationFileManager queued " + slowFetchers.size() + " slow files for fetching.");
        } else {
            LOG.debug("ExperimentSpecificationFileManager has no slow files to fetch.");
            eSpecLogger.firePostFileLoadAll(true);
        }

        //always shutdown, even if nothing was added.
        //   (otherwise, awaitTermination and isTerminated will not work as wanted)
        executorService.shutdown();

        LOG.debug("dir file sources: "+dirs.stream().map(FileSource::toString).collect(Collectors.joining(",")));
    }

    private void queueFileFetch(@Nonnull FileFetcher fileFetcher) {
        Runnable backGroundFetch = new Runnable() {
            @Override
            public void run() {
                try {
                    byte[] b = fileFetcher.fetchBytes();
                    eSpecLogger.firePostFileLoadSuccess(fileFetcher.getFileSource(), false, b.length);
                    bytes.put(fileFetcher.getFileSource(), b);
                } catch (IOException e) {
                    LOG.error("Failed to (slow) fetch file with source type "+
                            fileFetcher.getFileSource().getType()+"", e);
                    errorOccured = true;
                    eSpecLogger.firePostFileLoadFail(fileFetcher.getFileSource(), true, 0,
                            "Failed to (slow) fetch file with source type "+
                                    fileFetcher.getFileSource().getType()+"", e);
                    executorService.shutdownNow();
                    eSpecLogger.firePostFileLoadAll(false);
                }

                if (slowFetchCountdown.decrementAndGet() == 0 && !errorOccured) {
                    eSpecLogger.firePostFileLoadAll(true);
                }
            }
        };
        executorService.execute(backGroundFetch);
    }

    public boolean isReady() {
        if (!started)
            throw new IllegalStateException("Dangerous to call isReady before starting!");
        return executorService.isTerminated();
    }

    public boolean getErrorOccured() {
        return errorOccured;
    }

    /**
     * Wait for up to 30 minutes for all tasks to become ready
     *
     * @throws InterruptedException
     */
    public void waitUntilReady() throws InterruptedException {
        if (!started)
            throw new IllegalStateException("Dangerous to call waitUntilReady before starting!");

        if (isReady()) {
            LOG.debug("ExperimentSpecificationFileManager did not have to wait for all files to be ready");
            return;
        }

        LOG.debug("ExperimentSpecificationFileManager waits for all files to be ready");

        boolean terminated = this.executorService.awaitTermination(30, TimeUnit.MINUTES);
        if (!terminated)
            throw new IllegalStateException("Failed to wait until all files are ready");

        LOG.debug("ExperimentSpecificationFileManager stopped waiting for files to be ready ready");
    }

    @Nonnull
    public byte[] getBytes(@Nonnull FileSource fileSource) {
        if (!isReady())
            throw new IllegalStateException("ExperimentSpecificationFileManager requires you to wait until all files are ready, before using them.");

        byte[] res =  bytes.get(fileSource);
        if (res == null) {
            throw new IllegalStateException("Unexpectedly did not have content of " + fileSource.getBasename() + " (debug details: search fileSource="+fileSource+". Known fileSources="+bytes.keySet()+")");
        }
        return res;
    }

    public boolean isDir(@Nonnull FileSource fileSource) {
        boolean res = dirs.contains(fileSource);
        LOG.debug("isDir="+res+" for "+fileSource+" with dirs="+dirs.stream().map(FileSource::toString).collect(Collectors.joining(",")));
        return res;
    }

    public static class DirFileBytes {
        @Nonnull private final String baseFilename;     //filename without dirs
        @Nonnull private final String fullRelativePath;

        private final boolean dir;
        @Nonnull private final byte[] content;

        private DirFileBytes(@Nonnull String baseFilename, @Nonnull String fullRelativePath, @Nonnull byte[] content) {
            this.baseFilename = baseFilename;
            this.fullRelativePath = fullRelativePath;
            this.content = content;
            this.dir = false;
        }
        private DirFileBytes(@Nonnull String baseFilename, @Nonnull String fullRelativePath) {
            this.baseFilename = baseFilename;
            this.fullRelativePath = fullRelativePath;
            this.dir = true;
            this.content = new byte[0];
        }

        /**
         * special constructor for non dir files
         * @param content
         */
        public DirFileBytes(@Nonnull byte[] content) {
            this.baseFilename = "";
            this.fullRelativePath = "";
            this.content = content;
            this.dir = false;
        }

        @Nonnull
        public String getBaseFilename() {
            return baseFilename;
        }

        @Nonnull
        public String getFullRelativePath() {
            return fullRelativePath;
        }

        public boolean isDir() {
            return dir;
        }

        @Nonnull
        public byte[] getContent() {
            return content;
        }
    }

    private static class DirFile {
        @Nonnull String fullname;     //full name relative to bundle root
        @Nonnull String baseFilename; //filename without dirs
        public DirFile(@Nonnull String fullname) {
            this(fullname, IOUtils.getUnixBasename(fullname));
        }
        public DirFile(@Nonnull String fullname, @Nonnull String baseFilename) {
            this.fullname = fullname;
            this.baseFilename = baseFilename;

            if (fullname.isEmpty())
                throw new IllegalArgumentException("fullname is empty for baseFilename="+baseFilename);
        }
    }

    /**
     * This returns an iterator that returns all files and subdirs in the dir referenced in the fileSource
     *
     * For files, the content of the files can be requested.
     * Directories are guaranteed to be seen before any files in them.
     *
     * The first file returned is the dir itself, which has an empty basename and fullname
     *
     * @param fileSource
     * @return
     * @throws IOException
     */
    public Iterator<DirFileBytes> getDirBytes(@Nonnull FileSource fileSource) throws IOException {
        if (!isReady())
            throw new IllegalStateException("ExperimentSpecificationFileManager requires you to wait until all files are ready, before using them.");

        if (!isDir(fileSource))
            throw new IllegalStateException("Provided fileSource is not a directory");

        ESpecBundle bundleWithDir;
        String startDirPath;
        if (fileSource.getType() == FileSource.SourceType.GITHUB ||
                fileSource.getType() == FileSource.SourceType.GIT) {
            bundleWithDir = getCachedGitDir(fileSource);
            startDirPath = "/"; //root of bundle is already the requested subDir in the repo
        } else {
            bundleWithDir = bundle;
            startDirPath = assureEndWithSlash(fileSource.getValue());
        }

        if (fileSource.getType() != FileSource.SourceType.BUNDLED &&
                fileSource.getType() != FileSource.SourceType.GIT &&
                fileSource.getType() != FileSource.SourceType.GITHUB)
            throw new UnsupportedOperationException("Fetching files in directory not supported for file source "+fileSource.getType());


        DirFile dirItself = new DirFile(startDirPath, "");

        List<String> files = bundleWithDir.getDirFiles(startDirPath);

        LinkedList<DirFile> toCheck = files.stream()
                .map(filename -> new DirFile(filename))
                .collect(Collectors.toCollection(LinkedList::new));

        return new Iterator<DirFileBytes>() {
            LinkedList<DirFile> files = new LinkedList<>();
            LinkedList<DirFile> dirs = new LinkedList<>();

            {
                dirs.add(dirItself);
            }

            private void check() {
                while (files.isEmpty() && dirs.isEmpty() && !toCheck.isEmpty()) {
                    DirFile cur = toCheck.pop();
                    try {
                        if (bundleWithDir.isDir(cur.fullname)) {
                            toCheck.addAll(
                                    bundleWithDir.getDirFiles(cur.fullname).stream()
                                            .map(filename -> new DirFile(filename))
                                            .collect(Collectors.toList())
                            );
                            dirs.add(cur);
                        } else {
                            files.add(cur);
                        }
                    } catch (IOException e) {
                        throw new RuntimeException("Error checking next dir file", e);
                    }
                    if (LOG.isDebugEnabled()) dump("check()");
                }
            }

            @Override
            public boolean hasNext() {
                if (LOG.isDebugEnabled()) dump("hasNext()");
                if (files.isEmpty() && dirs.isEmpty() && !toCheck.isEmpty())
                    check();
                return !(toCheck.isEmpty() && files.isEmpty() && dirs.isEmpty());
            }

            @Override
            public DirFileBytes next() {
                if (LOG.isDebugEnabled()) dump("next()");
                if (!dirs.isEmpty()) {
                    //first dirs, if no dirs left, files.
                    DirFile n = dirs.pop();
                    return new DirFileBytes(
                            n.baseFilename,
                            removeParentDir(startDirPath, n.fullname));
                }

                DirFile n = files.pop();
                try {
                    return new DirFileBytes(
                            n.baseFilename,
                            removeParentDir(startDirPath, n.fullname),
                            bundleWithDir.getFileContent(n.fullname));
                } catch (IOException e) {
                    throw new RuntimeException("Error reading next dir file", e);
                }
            }

            private void dump(String prefix) {
                LOG.debug(" dirIt DUMP "+prefix
                        +" files="+files.stream().map(d -> d.fullname).collect(Collectors.toList())
                        +" toCheck="+toCheck.stream().map(d -> d.fullname).collect(Collectors.toList())
                );
            }
        };
    }

    private Map<FileSource, ESpecBundle> cachedGitDirBundles = new HashMap<>();
    private ESpecBundle getCachedGitDir(@Nonnull FileSource fileSource) {
        assert fileSource.getType() == FileSource.SourceType.GITHUB ||
                fileSource.getType() == FileSource.SourceType.GIT;
        GitFileSource gitFileSource = (GitFileSource) fileSource;
        assert !gitFileSource.isFile() : "Expected GitFileSource to not be a file but got: "+fileSource;
        assert gitFileSource.isDir() : "Expected GitFileSource to be a dir but got: "+fileSource;
        return cachedGitDirBundles.computeIfAbsent(fileSource, this::fetchGitDir);
    }

    @Nonnull
    private ESpecBundle fetchGitDir(@Nonnull FileSource fileSource) {
        assert fileSource.getType() == FileSource.SourceType.GITHUB ||
                fileSource.getType() == FileSource.SourceType.GIT;
        GitFileSource gitFileSource = (GitFileSource) fileSource;
        assert !gitFileSource.isFile();
        assert gitFileSource.isDir();
        assert gitFileSource.getFilename() == null;
        if (gitFileSource.isFile() || !gitFileSource.isDir() || gitFileSource.getFilename() != null)
            throw new IllegalArgumentException("Cannot fetch dir on a file");

        GitFetcherCommandBuilder commandBuilder = new GitFetcherCommandBuilder()
                .setGitUrl(gitFileSource.getGitUrl())
                .setBranch(gitFileSource.getBranch())
                .setRepoSubDir(gitFileSource.getRepoSubDir())
                .setUsername(gitFileSource.getUsername())
                .setPassword(gitFileSource.getPassword())
                .setPrivateKeyPemFile(gitFileSource.getPrivateKeyPem());
        GitFetcherCommand command = commandBuilder.build();

        GitFetcher fetcher = new GitFetcher(gitAuthPreferences);
        File localDir = fetcher.fetch(command);
        if (!localDir.exists())
            throw new RuntimeException("Target dir ("+localDir.getPath()+") does not exist in downloaded git repo");
        if (!localDir.isDirectory())
            throw new RuntimeException("Target dir ("+localDir.getPath()+") is not a dir in downloaded git repo");

        try {
            return BundleFetcher.fetchDir(localDir, false);
        } catch (ESpecBundle.ESpecBundleInitException e) {
            throw new RuntimeException("Error passing cloned git repo to BundleFetcher.fetchFir", e);
        }
    }


    @Nonnull
    public ExperimentSpecification getExperimentSpecification() {
        return eSpec;
    }

    @Nonnull
    public ESpecBundle getBundle() {
        return bundle;
    }



    @Nonnull
    private static String removeParentDir(@Nonnull String parentPath, @Nonnull String path) {
        if (parentPath.isEmpty()) return path;
        assert path.startsWith(parentPath) : "\""+path+"\" does not start with \""+parentPath+"\"";
        return path.substring(parentPath.length());
    }

    @Nonnull
    private static String assureEndWithSlash(@Nonnull String p) {
        if (p.endsWith("/")) return p;
        return p + "/";
    }

    @Nonnull
    private static String assureStartWithSlash(@Nonnull String p) {
        if (p.startsWith("/")) return p;
        return "/" + p;
    }
}
